From 98fcda92959d46a471e3b64e90dd74489c3d992a Mon Sep 17 00:00:00 2001
From: amit <amitlangote09@gmail.com>
Date: Fri, 10 May 2019 13:52:41 +0900
Subject: [PATCH] Fix bugs in pruning with composite range partition keys

* Force using equality (= operator) semantics when comparing clause
  expressions against partition bounds if the expression for the
  last matched key cannot be evaluated in a given pruning context;
  for example, if the expression is stable and pruning is invoked by
  the planner

* Ignore keys after a given key for which we detected a clause
  containing a non-inclusive operator
---
 src/backend/partitioning/partprune.c          | 88 +++++++++++++--------------
 src/test/regress/expected/partition_prune.out | 45 ++++++++++++++
 src/test/regress/sql/partition_prune.sql      |  8 +++
 3 files changed, 95 insertions(+), 46 deletions(-)

diff --git a/src/backend/partitioning/partprune.c b/src/backend/partitioning/partprune.c
index b5c1c7d4dd..7c894fc6ad 100644
--- a/src/backend/partitioning/partprune.c
+++ b/src/backend/partitioning/partprune.c
@@ -1152,17 +1152,24 @@ gen_prune_steps_from_opexps(PartitionScheme part_scheme,
 	List	   *opsteps = NIL;
 	List	   *btree_clauses[BTMaxStrategyNumber + 1],
 			   *hash_clauses[HTMaxStrategyNumber + 1];
-	bool		need_next_less,
-				need_next_eq,
-				need_next_greater;
+	bool		consider_next_key;
 	int			i;
 
+	consider_next_key = true;
 	memset(btree_clauses, 0, sizeof(btree_clauses));
 	memset(hash_clauses, 0, sizeof(hash_clauses));
 	for (i = 0; i < part_scheme->partnatts; i++)
 	{
 		List	   *clauselist = keyclauses[i];
-		bool		consider_next_key = true;
+
+		/*
+		 * If the previous key was found to be compared using such an operator
+		 * that comparisons of subsequent key columns against partition bounds
+		 * that occurs during partition pruning would be redundant, then don't
+		 * bother generating pruning steps for them.
+		 */
+		if (!consider_next_key)
+			break;
 
 		/*
 		 * To be useful for pruning, we must have clauses for a prefix of
@@ -1182,7 +1189,6 @@ gen_prune_steps_from_opexps(PartitionScheme part_scheme,
 			clauselist == NIL && !bms_is_member(i, nullkeys))
 			return NULL;
 
-		need_next_eq = need_next_less = need_next_greater = true;
 		foreach(lc, clauselist)
 		{
 			PartClauseInfo *pc = (PartClauseInfo *) lfirst(lc);
@@ -1198,13 +1204,13 @@ gen_prune_steps_from_opexps(PartitionScheme part_scheme,
 										   &lefttype,
 										   &righttype);
 
+
 			switch (part_scheme->strategy)
 			{
 				case PARTITION_STRATEGY_LIST:
 				case PARTITION_STRATEGY_RANGE:
 					{
 						PartClauseInfo *last = NULL;
-						bool		inclusive = false;
 
 						/*
 						 * Add this clause to the list of clauses to be used
@@ -1222,35 +1228,14 @@ gen_prune_steps_from_opexps(PartitionScheme part_scheme,
 								lappend(btree_clauses[pc->op_strategy], pc);
 
 						/*
-						 * We may not need the next clause if they're of
-						 * certain strategy.
+						 * We won't need the clause for the next key if the
+						 * clause for the current key contains a non-inclusive
+						 * operator.
 						 */
-						switch (pc->op_strategy)
-						{
-							case BTLessEqualStrategyNumber:
-								inclusive = true;
-								/* fall through */
-							case BTLessStrategyNumber:
-								if (!inclusive)
-									need_next_eq = need_next_less = false;
-								break;
-							case BTEqualStrategyNumber:
-								/* always accept clauses for the next key. */
-								break;
-							case BTGreaterEqualStrategyNumber:
-								inclusive = true;
-								/* fall through */
-							case BTGreaterStrategyNumber:
-								if (!inclusive)
-									need_next_eq = need_next_greater = false;
-								break;
-						}
+						if (pc->op_strategy == BTLessStrategyNumber ||
+							pc->op_strategy == BTGreaterStrategyNumber)
+							consider_next_key = false;
 
-						/* We may want to change our mind. */
-						if (consider_next_key)
-							consider_next_key = (need_next_eq ||
-												 need_next_less ||
-												 need_next_greater);
 						break;
 					}
 
@@ -1267,13 +1252,6 @@ gen_prune_steps_from_opexps(PartitionScheme part_scheme,
 					break;
 			}
 		}
-
-		/*
-		 * If we've decided that clauses for subsequent partition keys
-		 * wouldn't be useful for pruning, don't search any further.
-		 */
-		if (!consider_next_key)
-			break;
 	}
 
 	/*
@@ -2736,7 +2714,7 @@ get_matching_range_bounds(PartitionPruneContext *context,
 
 			/*
 			 * Look for the greatest bound that is < or <= lookup value and
-			 * set minoff to its offset.
+			 * set maxoff to its offset.
 			 */
 			off = partition_range_datum_bsearch(partsupfunc,
 												partcollation,
@@ -3103,11 +3081,29 @@ perform_pruning_base_step(PartitionPruneContext *context,
 											opstep->nullkeys);
 
 		case PARTITION_STRATEGY_RANGE:
-			return get_matching_range_bounds(context,
-											 opstep->opstrategy,
-											 values, nvalues,
-											 partsupfunc,
-											 opstep->nullkeys);
+			{
+				int		opstrategy = opstep->opstrategy;
+
+				/*
+				 * If we got values for only a subset (prefix) of the
+				 * expressions in opstep->exprs, adjust the strategy to
+				 * be used for pruning.  Originally, it would be the strategy
+				 * of the operator comparing the *last* expression against the
+				 * corresponding partition key, but it would be wrong to use
+				 * the same strategy now that we'll be comparing the values of
+				 * only a prefix of expressions against the partition bounds.
+				 * Use equality strategy, because all expressions in the
+				 * prefix must have been extracted from clauses containing
+				 * equality (=) or inclusive operators (<=, >=) anyway.
+				 */
+				if (nvalues < list_length(opstep->exprs))
+					opstrategy = BTEqualStrategyNumber;
+				return get_matching_range_bounds(context,
+												 opstrategy,
+												 values, nvalues,
+												 partsupfunc,
+												 opstep->nullkeys);
+			}
 
 		default:
 			elog(ERROR, "unexpected partition strategy: %d",
diff --git a/src/test/regress/expected/partition_prune.out b/src/test/regress/expected/partition_prune.out
index 6f320dfff4..889d4e14e2 100644
--- a/src/test/regress/expected/partition_prune.out
+++ b/src/test/regress/expected/partition_prune.out
@@ -938,6 +938,51 @@ explain (costs off) select * from mc3p where (a = 1 and abs(b) = 1) or (a = 10 a
          Filter: (((a = 1) AND (abs(b) = 1)) OR ((a = 10) AND (abs(b) = 9)))
 (9 rows)
 
+--
+-- Check that pruning with composite range partitioning works correctly
+-- correctly in the face of combination of static and runtime values being
+-- passed to perform plan-time pruning with
+--
+explain (analyze, costs off, summary off, timing off) select * from mc3p where a = 1 and abs(b) < (select 2);
+                       QUERY PLAN                       
+--------------------------------------------------------
+ Append (actual rows=0 loops=1)
+   InitPlan 1 (returns $0)
+     ->  Result (actual rows=1 loops=1)
+   ->  Seq Scan on mc3p0 (actual rows=0 loops=1)
+         Filter: ((a = 1) AND (abs(b) < $0))
+   ->  Seq Scan on mc3p1 (actual rows=0 loops=1)
+         Filter: ((a = 1) AND (abs(b) < $0))
+   ->  Seq Scan on mc3p_default (actual rows=0 loops=1)
+         Filter: ((a = 1) AND (abs(b) < $0))
+(9 rows)
+
+explain (analyze, costs off, summary off, timing off) select * from mc3p where a < (select 1) and abs(b) = 1;
+                       QUERY PLAN                       
+--------------------------------------------------------
+ Append (actual rows=0 loops=1)
+   InitPlan 1 (returns $0)
+     ->  Result (actual rows=1 loops=1)
+   ->  Seq Scan on mc3p0 (actual rows=0 loops=1)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p1 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p2 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p3 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p4 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p5 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p6 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p7 (never executed)
+         Filter: ((a < $0) AND (abs(b) = 1))
+   ->  Seq Scan on mc3p_default (actual rows=0 loops=1)
+         Filter: ((a < $0) AND (abs(b) = 1))
+(21 rows)
+
 -- a simpler multi-column keys case
 create table mc2p (a int, b int) partition by range (a, b);
 create table mc2p_default partition of mc2p default;
diff --git a/src/test/regress/sql/partition_prune.sql b/src/test/regress/sql/partition_prune.sql
index eafbec6f35..b78b08dfc1 100644
--- a/src/test/regress/sql/partition_prune.sql
+++ b/src/test/regress/sql/partition_prune.sql
@@ -122,6 +122,14 @@ explain (costs off) select * from mc3p where a = 1 or abs(b) = 1 or c = 1;
 explain (costs off) select * from mc3p where (a = 1 and abs(b) = 1) or (a = 10 and abs(b) = 10);
 explain (costs off) select * from mc3p where (a = 1 and abs(b) = 1) or (a = 10 and abs(b) = 9);
 
+--
+-- Check that pruning with composite range partitioning works correctly
+-- correctly in the face of combination of static and runtime values being
+-- passed to perform plan-time pruning with
+--
+explain (analyze, costs off, summary off, timing off) select * from mc3p where a = 1 and abs(b) < (select 2);
+explain (analyze, costs off, summary off, timing off) select * from mc3p where a < (select 1) and abs(b) = 1;
+
 -- a simpler multi-column keys case
 create table mc2p (a int, b int) partition by range (a, b);
 create table mc2p_default partition of mc2p default;
-- 
2.11.0

